import os
import argparse
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick

# Plot format
fs=8
plt.rc('font', family='serif')
plt.rc('xtick', labelsize='x-small')
plt.rc('ytick', labelsize='x-small')
plt.rc('text', usetex=True)

def plot_log(log_file,out_dir,fig_format='png',use_dark_mode=False):
  log_data = np.loadtxt(log_file)

  # Set dark mode
  if(use_dark_mode):
    plt.style.use('dark_background')  

  # loss profile
  plt.figure(figsize=(2,2))
  plt.semilogy(log_data[:,1],log_data[:,2],'b-')
  plt.xlabel('Iterations',fontsize=fs)
  plt.ylabel('log loss',fontsize=fs)
  plt.gca().tick_params(axis='both', labelsize=fs)
  plt.tight_layout()
  plt.savefig(out_dir+'log_plot.'+fig_format,bbox_inches='tight',dpi=200)
  plt.close()

def plot_params(param_data,LL_data,idx1,idx2,out_dir,out_info,fig_format='png',use_dark_mode=False):  

  # Read data
  param_data = np.loadtxt(param_data)  
  dent_data  = np.loadtxt(LL_data)

  # Set dark mode
  if(use_dark_mode):
    plt.style.use('dark_background')

  # Plot figure
  plt.figure(figsize=(3,2))
  plt.scatter(param_data[:,idx1],param_data[:,idx2],s=1.5,lw=0,marker='o',c=np.exp(dent_data))
  plt.colorbar()
  plt.gca().xaxis.set_major_formatter(mtick.FormatStrFormatter('%.1f'))
  plt.gca().yaxis.set_major_formatter(mtick.FormatStrFormatter('%.1e'))
  plt.gca().tick_params(axis='both', labelsize=fs)
  plt.xlabel('$z_{K,'+str(idx1+1)+'}$',fontsize=fs)
  plt.ylabel('$z_{K,'+str(idx2+1)+'}$',fontsize=fs)  
  # Set limits based on avg and std
  avg_1 = np.mean(param_data[:,idx1])
  std_1 = np.std(param_data[:,idx1])
  avg_2 = np.mean(param_data[:,idx2])
  std_2 = np.std(param_data[:,idx2])  
  plt.xlim([avg_1-3*std_1,avg_1+3*std_1])
  plt.ylim([avg_2-3*std_2,avg_2+3*std_2])
  plt.tight_layout()
  plt.savefig(out_dir+'params_plot_' + out_info + '_'+str(idx1)+'_'+str(idx2)+'.'+fig_format,bbox_inches='tight',dpi=200)
  plt.close()

def plot_outputs(sample_file,obs_file,idx1,idx2,out_dir,out_info,fig_format='png',use_dark_mode=False):  

  # Read data
  sample_data = np.loadtxt(sample_file)  
  obs_data    = np.loadtxt(obs_file) 

    # Set dark mode
  if(use_dark_mode):
    plt.style.use('dark_background')

  plt.figure(figsize=(2.5,2))
  plt.scatter(sample_data[:,idx1],sample_data[:,idx2],s=2,c='b',marker='o',edgecolor=None,alpha=0.1)
  plt.scatter(obs_data[idx1,:],obs_data[idx2,:],s=3,c='r',alpha=1,zorder=99)
  plt.gca().xaxis.set_major_formatter(mtick.FormatStrFormatter('%.2f'))
  plt.gca().yaxis.set_major_formatter(mtick.FormatStrFormatter('%.2f'))
  plt.gca().tick_params(axis='both', labelsize=fs)
  plt.xlabel('$x_{'+str(idx1+1)+'}$',fontsize=fs)
  plt.ylabel('$x_{'+str(idx2+1)+'}$',fontsize=fs)  
  # Set limits based on avg and std
  avg_1 = np.mean(sample_data[:,idx1])
  std_1 = np.std(sample_data[:,idx1])
  avg_2 = np.mean(sample_data[:,idx2])
  std_2 = np.std(sample_data[:,idx2])  
  plt.xlim([avg_1-3*std_1,avg_1+3*std_1])
  plt.ylim([avg_2-3*std_2,avg_2+3*std_2])
  plt.tight_layout()
  plt.savefig(out_dir+'data_plot_' + out_info + '_'+str(idx1)+'_'+str(idx2)+'.'+fig_format,bbox_inches='tight',dpi=200)
  plt.close()

# =========
# MAIN CODE
# =========
if __name__ == '__main__':  

  # Init parser
  parser = argparse.ArgumentParser(description='.')

  # folder name
  parser.add_argument('-f', '--folder',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      required=False,
                      help='Folder with experiment results',
                      metavar='',
                      dest='folder_name')


  # folder name
  parser.add_argument('-n', '--name',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      required=True,
                      help='Name of numerical experiment',
                      metavar='',
                      dest='exp_name')

  # iteration number = 1
  parser.add_argument('-i', '--iter',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=1,
                      type=int,
                      choices=None,
                      required=True,
                      help='Iteration number',
                      metavar='',
                      dest='step_num')
  
  # plot format
  parser.add_argument('-p', '--picformat',
                      action=None,
                      const=None,
                      default='png',
                      type=str,
                      choices=['png','pdf','jpg'],
                      required=False,
                      help='Output format for picture',
                      metavar='',
                      dest='img_format')

  # Enable dark mode for pictures
  parser.add_argument('-d', '--dark',
                      action='store_true',
                      default=False,
                      required=False,
                      help='Generate pictures for dark background',
                      dest='use_dark_mode')

  # Parse Commandline Arguments
  args = parser.parse_args()

  # Set file name/path
  out_dir     = args.folder_name + args.exp_name + '/'
  log_file    = out_dir + 'log.txt'
  sample_file = out_dir + args.exp_name + '_samples_'    + str(args.step_num)
  param_file  = out_dir + args.exp_name + '_params_'     + str(args.step_num)
  LL_file     = out_dir + args.exp_name + '_logdensity_' + str(args.step_num)
  output_file = out_dir + args.exp_name + '_outputs_'    + str(args.step_num)
  obs_file    = out_dir + args.exp_name + '_data'
  out_info    = args.exp_name + '_' + str(args.step_num)

  # Plot loss profile
  if(os.path.isfile(log_file)):
    print('Plotting log...')
    plot_log(log_file,out_dir,fig_format=args.img_format,use_dark_mode=args.use_dark_mode)
  else:
    print('Log file not found: '+log_file)

  # Plot 2D slice of posterior samples
  if(os.path.isfile(param_file) and os.path.isfile(LL_file)):
    tot_params  = np.loadtxt(param_file).shape[1]
    print('Plotting posterior samples...')
    for loopA in range(tot_params):
      for loopB in range(loopA+1, tot_params):
        plot_params(param_file,LL_file,loopA,loopB,out_dir,out_info,fig_format=args.img_format,use_dark_mode=args.use_dark_mode)
  else:
    print('File with posterior samples not found: '+param_file)
    print('File with log-density not found: '+LL_file)

  # Plot 2D slice of outputs and observations
  if(os.path.isfile(output_file) and os.path.isfile(obs_file)):
    tot_outputs = np.loadtxt(output_file).shape[1]
    print('Plotting posterior predictive samples...')
    for loopA in range(tot_outputs):
      for loopB in range(loopA+1, tot_outputs):
        plot_outputs(output_file,obs_file,loopA,loopB,out_dir,out_info,fig_format=args.img_format,use_dark_mode=args.use_dark_mode)
  else:
    print('File with posterior predictive samples not found: '+output_file)
    print('File with observations not found: '+obs_file)



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from os import path

torch.set_default_tensor_type(torch.DoubleTensor)

class FNN(nn.Module):
    """Fully Connected Neural Network"""

    def __init__(self, input_size, output_size, arch=None, activation='relu', device='cpu'):
        """
        Args:
            input_size (int): Input size for FNN
            output_size (int): Output size for FNN
            arch (list of int): list containing the number of neurons for each hidden layer
            activation (stirng): type of activation function, either 'relu', 
            device (string): computational device
        """
        super().__init__()
        if(arch is None):
            self.neuron_size = [64,32]
        else:
            self.neuron_size = arch
        # Set activation
        self.activation = activation
        self.fc = nn.ModuleList()
        # Assign first layer
        self.fc.append(nn.Linear(input_size, self.neuron_size[0]).to(device))
        if(len(self.neuron_size) == 1):
            # Assign output Layer
            self.fc.append(nn.Linear(self.neuron_size[0], output_size).to(device))
        else:
            for layer in range(1,len(self.neuron_size)):
                self.fc.append(nn.Linear(self.neuron_size[layer-1], self.neuron_size[layer]).to(device))
            # Assign output layer
            self.fc.append(nn.Linear(self.neuron_size[len(self.neuron_size)-1], output_size).to(device))

    def forward(self, x):
        """
        Args:
            x (torch.Tensor): [num_samples, input_size], assumed to be a batch.

        Returns:
            torch.Tensor. Assumed to be a batch.
        """
        for loopA in range(len(self.fc)-1):
            if(self.activation == 'relu'):
                x = F.relu(self.fc[loopA](x))
            elif(self.activation == 'silu'):
                x = F.silu(self.fc[loopA](x))
            elif(self.activation == 'tanh'):
                x = F.tanh(self.fc[loopA](x))
            else:
                print('Invalid activation string.')
                exit(-1)

        # Last layer with linear activation
        x = self.fc[len(self.fc)-1](x)
        return x

class Surrogate:
    """Class to create surrogate models for inference with NoFAS

       Args:
            model_name (string): name of the true model to be approximated
            model_func (function): with only one input of torch.Tensor
            input_size (int): input dimension of true model.
            output_size (int): output dimension of true model.
            dnn_arch (list of int): list containing the number of neurons for each hidden layer. Default None for [64,32] architecture.
            dnn_activation (stirng): type of activation function, either 'relu', 
            model_folder (string): folder where surrogate model is stored.             
            limits (list or lists): bounds for all inputs, in format of [[low_0, high_0], [low_1, high_1], ... ]
            memory_len (int): the maximal number of batches stored in buffer. Default: 20
            surrogate (None or torch.nn.Module): the implementation of surrogate model used. Default: FNN This allows the user to spacify any NN architecture for the surrogate.
                                                 If None the parameters "dnn_arch" and "dnn_activation" can be used to specify a dense neural network. 
    """
    def __init__(self, model_name, model_func, input_size, output_size, dnn_arch=None, dnn_activation='relu', model_folder='./', limits=None, memory_len=20, surrogate=None, device='cpu'):
        self.device = device
        self.input_size = input_size
        self.output_size = output_size
        self.model_name = model_name
        self.model_folder = model_folder
        self.mf = model_func
        self.pre_out = None
        self.m = None
        self.sd = None
        self.tm = None
        self.tsd = None
        self.limits = limits
        self.pre_grid = None
        self.surrogate = FNN(input_size, output_size, arch=dnn_arch, device=self.device) if surrogate is None else surrogate
        self.beta_0 = 0.5
        self.beta_1 = 0.1        

        self.memory_grid = []
        self.memory_out = []
        self.memory_len = memory_len
        self.weights = torch.Tensor([np.exp(-self.beta_1 * i) for i in range(memory_len)]).to(self.device)
        self.grid_record = None
        

    @property
    def limits(self):
        """Retrieves the limits of all inputs
        """
        return self.__limits

    @limits.setter
    def limits(self, limits):
        """Sets the limits of all inputs
        
        Args:
            limits (list or tuple): Lists lower and upper bound in the format *[[low_0, high_0], [low_1, high_1], ...]*

        """
        limits = torch.Tensor(limits).tolist()
        if len(limits) != self.input_size:
            print("Error: Invalid input size for limit. Abort.")
            exit(-1)
        elif any(len(item) != 2 for item in limits):
            print("Error: Limits should be two bounds. Abort.")
            exit(-1)
        elif any(item[0] > item[1] for item in limits):
            print("Error: Upper bound should not be smaller than lower bound. Abort.")
            exit(-1)
        self.__limits = limits

    @property
    def pre_grid(self):
        return self.__pre_grid

    @pre_grid.setter
    def pre_grid(self, pre_grid):
        """Assign the pregrid for the surrogate model. 

        f nofas is not enabled, this serves as the whole training grid.
        
        Args:
            pre_grid (None or torch.Tensor): Specifies a matrix of model inputs with size [data_num, feature_dim]. If None, will try to search for *npz* containing this info.

        Returns:
            None

        """
        if pre_grid is None:
            if path.exists(self.model_folder + self.model_name + '.npz'):
                container = np.load(self.model_folder + self.model_name + '.npz')
                if 'pre_grid' in container:
                    self.pre_grid = container['pre_grid']
                    print("Success: Pre-Grid found.")
            else:
                print("Warning: " + self.model_folder + self.model_name + ".npz does not found, please generate pre-grid.")
                print("Suggestion: Use Surrogate.gen_grid(input_limits=None, grid_num=5, store=True)")
        else:
            self.__pre_grid = torch.Tensor(pre_grid).to(self.device)
            self.m = torch.mean(self.pre_grid, 0)
            self.sd = torch.std(self.pre_grid, 0)
            # Evaluate model at pre-grid
            self.pre_out = self.mf(self.pre_grid)
            self.tm = torch.mean(self.pre_out, 0)
            self.tsd = torch.std(self.pre_out, 0)
            self.grid_record = self.__pre_grid.clone()

    def gen_grid(self, input_limits=None, grid_type='tensor', gridnum=4, store=True):
        """Generates a pre-grid.
        
        Args:
            input_limits (None or list[list]): If None, use self.limits. If list[list], rewrite self.limits and use it.
            gridnum (int): Contains the number of grid points per dimension.
            store (bool): Flag indicating the pre-grid is store in self.pre_grid. If False then self.pre_grid is None.

        Returns:
            torch.Tensor: Input values for the full tensor grid in *dim* dimensions stored in a matrix [gridnum ** dim, dim].
        """
        if (grid_type == 'tensor'):
          
            meshpoints = []
            if input_limits is not None:
                self.limits = input_limits
                print("Warning: Input limits recorded in surrogate.")

            for lim in self.limits: meshpoints.append(torch.linspace(lim[0], lim[1], steps=gridnum))
            grid = torch.meshgrid(meshpoints,indexing='ij')
            grid = torch.cat([item.reshape(gridnum ** len(self.limits), 1) for item in grid], 1)

        elif (grid_type == 'sobol'):

            # Generate sobol samples in [0,1]^d
            soboleng = torch.quasirandom.SobolEngine(dimension=len(self.limits))
            grid = soboleng.draw(gridnum)
            for i,lim in enumerate(self.limits): 
                grid[:,i] = lim[0] + (lim[1] - lim[0]) * grid[:,i]

        else:

            print('Invalid type for pre-grid generation')
            exit(-1)

        # Store pre-grid
        if store:
            self.pre_grid = grid
            self.grid_record = self.pre_grid.clone()
            self.surrogate_save()
        
        # Return grid
        return grid

    def surrogate_save(self):
        """Save surrogate model to [self.name].sur and [self.name].npz
        
        Returns:
            None

        """
        torch.save(self.surrogate.state_dict(), self.model_folder + self.model_name + '.sur')
        np.savez(self.model_folder + self.model_name, limits=self.limits, pre_grid=self.pre_grid.clone().cpu().numpy(),
                 grid_record=self.grid_record.clone().cpu().numpy())

    def surrogate_load(self):
        """Load surrogate model from [self.name].sur and [self.name].npz
        
        Returns:
            None
        """
        self.surrogate.load_state_dict(torch.load(self.model_folder + self.model_name + '.sur'))
        container = np.load(self.model_folder + self.model_name + '.npz')
        for key in container:
            try:
                setattr(self, key, torch.Tensor(container[key]))
                print("Success: [" + key + "] loaded.")
            except:
                print("Warning: [" + key + "] is not a surrogate variables.")

    def pre_train(self, max_iters, lr, lr_exp, record_interval, store=True, reg=False):
        """Train surrogate model with pre-grid.

         Training is performed with the RMSprop optimizer and with an exponential learning rate scheduler.

        Args:
            max_iters (int): The maximal number of iterations.
            lr (double): Learning rate for RMSprop.
            lr_exp (double): Decay factor for exponential scheduler.
            record_interval (int): The number of iterations to print loss info
            store (bool): If true, self.surrogate_save() will be called.
            reg (bool): If true, L1 regularization will be used with parameter 0.0001

        Returns:
            None
        """
        print('')
        print('--- Pre-training surrogate model')
        print('')
        grid = (self.pre_grid - self.m) / self.sd
        out = (self.pre_out - self.tm) / self.tsd
        optimizer = torch.optim.RMSprop(self.surrogate.parameters(), lr=lr)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, lr_exp)
        for i in range(max_iters):
            self.surrogate.train()            
            y = self.surrogate(grid)
            loss = torch.sum((y - out) ** 2) / y.size(0)
            if reg:
                reg_loss = 0
                for param in self.surrogate.parameters():
                    reg_loss += torch.abs(param).sum() * 0.0001
                loss += reg_loss
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            scheduler.step()

            if i % record_interval == 0:
                if reg:
                    print('SUR: PRE: it: %7d | loss: %8.3e | reg_loss: %8.3e' % (i, loss, reg_loss))
                else:
                    print('SUR: PRE: it: %7d | loss: %8.3e' % (i, loss))
        print('')
        print('--- Surrogate model pre-train complete')
        print('')
        if store:
            self.surrogate_save()

    def update(self, x, max_iters=10000, lr=0.01, lr_exp=0.999, record_interval=500, store=False, tol=1e-5, reg=False):
        """Model calibration for NoFAS.
        
        Fine tuning is performed with the RMSprop optimizer and an exponential learning rate scheduler.
        
        Args:
            x (torch.Tensor): Input feature that needs true model output.
            max_iters (int): Maximum number of iteration. Default 10000.
            lr (double): Learning rate for RMSprop. Default 0.01.
            lr_exp (double): Decay factor of exponential scheduler.
            record_interval (int): The number of iterations to print loss information.
            store (bool): If ture, self.surrogate_save() will be called.
            tol (double): Optimization will be terminated if loss < tol ** 2.
            reg (bool): If ture, L1 regularizaiton will be used with parameter 0.1.

        Returns:
            None
        """
        self.grid_record = torch.cat((self.grid_record.to(self.device), x), dim=0)
        s = torch.std(x, dim=0)
        thresh = torch.tensor(0.1).to(self.device)
        if torch.any(s < thresh):
            p = x[:, s < thresh]
            x[:, s < thresh] += torch.normal(0, 1, size=tuple(p.size())).to(self.device) * thresh
        s_aft = torch.std(x, dim=0)            
        
        # Print the std for each dimension before and after inflation    
        print('')
        print('--- Updating surrogate model')
        print('')
        print('Std before inflation -> Std after inflation')
        for loopA in range(s.size(0)):
          print("%8.3e -> %8.3e" % (s[loopA],s_aft[loopA]))
        print('')          

        if len(self.memory_grid) >= self.memory_len:
            self.memory_grid.pop()
            self.memory_out.pop()
        self.memory_grid.insert(0, (x - self.m) / self.sd)
        self.memory_out.insert(0, (self.mf(x) - self.tm) / self.tsd)
        sizes = [list(self.pre_grid.size())[0]] + [list(item.size())[0] for item in self.memory_grid]

        optimizer = torch.optim.RMSprop(self.surrogate.parameters(), lr=lr)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, lr_exp)
        for i in range(max_iters):
            self.surrogate.train()            
            y = self.surrogate(torch.cat(((self.pre_grid - self.m) / self.sd, *self.memory_grid), dim=0))
            out = torch.cat(((self.pre_out - self.tm) / self.tsd, *self.memory_out), dim=0)
            raw_loss = torch.stack([item.mean() for item in torch.split(torch.sum((y - out) ** 2, dim=1), sizes)])
            loss = raw_loss[0] * 2 * self.beta_0 * self.weights[:len(self.memory_grid)].sum() + \
                   torch.sum(raw_loss[1:] * self.weights[:len(self.memory_grid)]) * (1 - self.beta_0) * 2

            # loss = raw_loss[0] * self.weights[:len(self.memory_grid)].sum() + torch.sum(
            #     raw_loss[1:] * self.weights[:len(self.memory_grid)])

            if reg:
                for param in self.surrogate.parameters():
                    loss += torch.abs(param).sum() * 0.1
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            scheduler.step()

            if i % record_interval == 0:
                # print('Updating: {}\t loss {}'.format(i, loss), end='\r')
                print('SUR: UPD: it: %7d | loss: %8.3e' % (i, loss))
            if loss < tol ** 2: break
        # print('                                                        ', end='\r')
        print('')
        print('--- Surrogate model updated')
        print('')                  
        if store:
            self.surrogate_save()

    def forward(self, x):
        """Function to evaluate the surrogate
        
        Args:
            x (torch.Tensor): Contains a matrix of model inputs in the form [data_num, feature_dim]

        Returns:
            Value of the surrogate at x.

        """
        return self.surrogate((x - self.m) / self.sd) * self.tsd + self.tm

def test_surrogate():
    
    import matplotlib.pyplot as plt
  
    # Define function to approximate
    def ishigami(z):
        z1, z2, z3 = torch.chunk(z, chunks=3, dim=1)
        return torch.sin(z1) + 7.0 * (torch.sin(z2))**2 + 0.1 * z3**4 * torch.sin(z1)
    
    # Set surrogate architecture
    # arch=[64,64,64,64]
    arch=None
    activation='silu'
    
    # Define emulator and pre-train on global grid
    curr_limits = [[-torch.pi,torch.pi],[-torch.pi,torch.pi],[-torch.pi,torch.pi]]
    emulator    = Surrogate(model_name='ishigami', model_func=ishigami, input_size=3, output_size=1, limits = curr_limits, 
                            dnn_arch=arch, dnn_activation=activation)
    emulator.gen_grid(gridnum=10)
    emulator.pre_train(40000, 0.03, 0.9999, 500, store=True)

    # Eval MSE on Sobol point collection
    soboleng = torch.quasirandom.SobolEngine(dimension=3)
    # 2**10 = 1024 testing points
    z = (soboleng.draw_base2(10) - 0.5) * 2 * torch.pi
    true_fc = ishigami(z)
    emul_fc = emulator.forward(z)
    mse = torch.sum((true_fc - emul_fc) ** 2) / z.size(0)
    print('Test MSE: %.3f' % (mse.item()))

    # Check loss on pregrid
    true_fc_pg = ishigami(emulator.pre_grid)
    emul_fc_pg = emulator.forward(emulator.pre_grid)
    mse = torch.sum((true_fc_pg - emul_fc_pg) ** 2) / emulator.pre_grid.size(0)
    print('Pre-grid MSE: %.3f' % (mse.item()))
    print('Standard deviation on testing set: %.3f' % (emulator.tsd.item()))

        # Plot output histograms
    plt.figure(figsize=(5,5))
    plt.hist(true_fc.detach().numpy(),bins='auto',alpha=0.5,density=True)
    plt.hist(emul_fc.detach().numpy(),bins='auto',alpha=0.5,density=True)
    plt.show()

# TEST SURROGATE
if __name__ == '__main__':
  
  test_surrogate()
    


import unittest
import os
import linfa
from linfa.run_experiment import experiment
from linfa.transform import Transformation
from linfa.nofas import Surrogate
import torch
import random
import numpy as np

class linfa_test_suite(unittest.TestCase):

    def trivial_example(self):

        if "it" in os.environ:
          max_it  = int(os.environ["it"])
          max_pre = 1000
        else:
          max_it  = 25001
          max_pre = 40000

        print('')
        print('--- TEST 1: TRIVIAL FUNCTION - NOFAS')
        print('')

        # Import trivial model
        from linfa.models.TrivialModels import Trivial

        exp = experiment()
        exp.name              = "trivial"
        exp.flow_type         = 'realnvp'     # str: Type of flow (default 'realnvp')
        exp.n_blocks          = 5             # int: Number of layers (default 5)
        exp.hidden_size       = 100           # int: Hidden layer size for MADE in each layer (default 100)
        exp.n_hidden          = 1             # int: Number of hidden layers in each MADE (default 1)
        exp.activation_fn     = 'relu'        # str: Actication function used (default 'relu')
        exp.input_order       = 'sequential'  # str: Input order for create_mask (default 'sequential')
        exp.batch_norm_order  = True          # bool: Order to decide if batch_norm is used
        exp.save_interval = 5000          # int: How often to sample from normalizing flow

        exp.input_size    = 2       # int: Dimensionality of input (default 2)
        exp.batch_size    = 200     # int: Number of samples generated (default 100)
        exp.true_data_num = 2       # double: number of true model evaluated (default 2)
        # exp.n_iter      = 25001
        exp.n_iter        = max_it  # int: Number of iterations (default 25001)
        exp.lr            = 0.002   # float: Learning rate (default 0.003)
        exp.lr_decay      = 0.9999  # float: Learning rate decay (default 0.9999)
        exp.log_interval  = 10      # int: How often to show loss stat (default 10)

        exp.run_nofas          = True
        exp.surr_pre_it        = max_pre #:int: Number of pre-training iterations for surrogate model
        exp.surr_upd_it        = 6000  #:int: Number of iterations for the surrogate model update
        exp.annealing          = False
        exp.calibrate_interval = 1000  # int: How often to update surrogate model (default 1000)
        exp.budget             = 64    # int: Total number of true model evaluation
        exp.surr_folder        = "./"
        exp.use_new_surr       = True

        exp.output_dir   = './results/' + exp.name
        exp.log_file     = 'log.txt'
        exp.seed         = random.randint(0, 10 ** 9)  # int: Random seed used
        exp.n_sample     = 5000                        # int: Batch size to generate final results/plots
        exp.no_cuda      = True                        # Running on CPU by default but tested on CUDA

        exp.optimizer    = 'RMSprop'
        exp.lr_scheduler = 'ExponentialLR'

        exp.device = torch.device('cuda:0' if torch.cuda.is_available() and not exp.no_cuda else 'cpu')

        # Define transformation
        # One list for each variable
        trsf_info = [['identity',0,0,0,0],
                     ['identity',0,0,0,0]]
        trsf = Transformation(trsf_info)        
        exp.transform = trsf

        # Define model
        model = Trivial(device=exp.device)
        exp.model = model

        # Get data
        res_path = os.path.abspath(os.path.join(os.path.dirname(linfa.__file__), os.pardir))
        model.data = np.loadtxt(res_path + '/resource/data/data_trivial.txt')

        # Define surrogate
        exp.surrogate = Surrogate(exp.name, lambda x: model.solve_t(trsf.forward(x)), 2, 2, 
                                  model_folder=exp.surr_folder, limits=[[0, 6], [0, 6]], 
                                  memory_len=20, device=exp.device)
        surr_filename = exp.surr_folder + exp.name
        if exp.use_new_surr or (not os.path.isfile(surr_filename + ".sur")) or (not os.path.isfile(surr_filename + ".npz")):
            print("Warning: Surrogate model files: {0}.npz and {0}.npz could not be found. ".format(surr_filename))
            # 4 samples for each dimension: pre-grid size = 16
            exp.surrogate.gen_grid(gridnum=4)
            exp.surrogate.pre_train(exp.surr_pre_it, 0.03, 0.9999, 500, store=True)
        # Load the surrogate
        exp.surrogate.surrogate_load()

        # Define log density
        def log_density(x, model, surrogate, transform):
            # x contains the original, untransformed inputs

            # Compute transformation log Jacobian
            adjust = transform.compute_log_jacob_func(x)

            # Eval model output
            stds = torch.abs(model.solve_t(model.defParam)) * model.stdRatio
            Data = torch.tensor(model.data).to(exp.device)
            if surrogate:
              modelOut = exp.surrogate.forward(x)
            else:
              modelOut = model.solve_t(transform.forward(x))
              
            # Eval LL
            ll1 = -0.5 * np.prod(model.data.shape) * np.log(2.0 * np.pi)
            ll2 = (-0.5 * model.data.shape[1] * torch.log(torch.prod(stds))).item()
            ll3 = 0.0
            for i in range(2):
              ll3 += - 0.5 * torch.sum(((modelOut[:, i].unsqueeze(1) - Data[i, :].unsqueeze(0)) / stds[0, i]) ** 2, dim=1)
            negLL = -(ll1 + ll2 + ll3)

            # Return LL
            return -negLL.reshape(x.size(0), 1) + adjust

        # Assign log-density model
        exp.model_logdensity = lambda x: log_density(x, model, exp.surrogate, trsf)

        # Run VI
        exp.run()


    def highdim_example(self):

        if "it" in os.environ:
          max_it  = int(os.environ["it"])
          max_pre = 1000
        else:
          max_it  = 25001
          max_pre = 100000

        print('')
        print('--- TEST 2: HIGH DIMENSIONAL SOBOL FUNCTION - NOFAS')
        print('')

        from linfa.models.highdimModels import Highdim
        
        exp = experiment()
        exp.name = "highdim"
        exp.flow_type         = 'realnvp'     # str: Type of flow (default 'realnvp')
        exp.n_blocks          = 15            # int: Number of layers (default 5)
        exp.hidden_size       = 100           # int: Hidden layer size for MADE in each layer (default 100)
        exp.n_hidden          = 1             # int: Number of hidden layers in each MADE (default 1)
        exp.activation_fn     = 'relu'        # str: Actication function used (default 'relu')
        exp.input_order       = 'sequential'  # str: Input order for create_mask (default 'sequential')
        exp.batch_norm_order  = True          # bool: Order to decide if batch_norm is used (default True)
        exp.save_interval     = 5000          # int: How often to sample from normalizing flow

        exp.input_size    = 5       # int: Dimensionality of input (default 2)
        exp.batch_size    = 250     # int: Number of samples generated (default 100)
        exp.true_data_num = 12      # double: number of true model evaluated (default 2)
        # exp.n_iter      = 25001
        exp.n_iter        = max_it  # int: Number of iterations (default 25001)
        exp.lr            = 0.003   # float: Learning rate (default 0.003)
        exp.lr_decay      = 0.9999  # float: Learning rate decay (default 0.9999)
        exp.log_interval  = 10      # int: How often to show loss stat (default 10)

        exp.run_nofas          = True
        exp.surr_pre_it       = max_pre #:int: Number of pre-training iterations for surrogate model
        exp.surr_upd_it       = 6000   #:int: Number of iterations for the surrogate model update

        exp.annealing          = False
        exp.calibrate_interval = 200   # int: How often to update surrogate model (default 1000)
        exp.budget             = 1023  # int: Total number of true model evaluation
        exp.surr_folder        = "./"
        exp.use_new_surr       = True

        exp.output_dir   = './results/' + exp.name
        exp.log_file     = 'log.txt'
        exp.seed         = random.randint(0, 10 ** 9)  # int: Random seed used
        exp.n_sample     = 5000                        # int: Batch size to generate final results/plots
        exp.no_cuda      = True                        # Running on CPU by default but tested on CUDA

        exp.optimizer    = 'RMSprop'
        exp.lr_scheduler = 'ExponentialLR'

        exp.device = torch.device('cuda:0' if torch.cuda.is_available() and not exp.no_cuda else 'cpu')

        # Define transformation
        # One list for each variable
        trsf_info = [['identity',0,0,0,0],
                     ['identity',0,0,0,0],
                     ['identity',0,0,0,0],
                     ['identity',0,0,0,0],
                     ['identity',0,0,0,0]]
        trsf = Transformation(trsf_info)
        exp.transform = trsf

        # Define the model
        model = Highdim(exp.device)
        exp.model = model

        # Read data
        res_path = os.path.abspath(os.path.join(os.path.dirname(linfa.__file__), os.pardir))
        model.data = np.loadtxt(res_path + '/resource/data/data_highdim.txt')

        # Define the surrogate
        exp.surrogate = Surrogate(exp.name, lambda x: model.solve_t(trsf.forward(x)), model.input_num, model.output_num,
                                  model_folder=exp.surr_folder, limits=torch.Tensor([[-3.0, 3.0], [-3.0, 3.0], [-3.0, 3.0], [-3.0, 3.0], [-3.0, 3.0]]), 
                                  memory_len=20, device=exp.device)
        surr_filename = exp.surr_folder + exp.name
        if exp.use_new_surr or (not os.path.isfile(surr_filename + ".sur")) or (not os.path.isfile(surr_filename + ".npz")):
            print("Warning: Surrogate model files: {0}.npz and {0}.npz could not be found. ".format(surr_filename))
            exp.surrogate.gen_grid(gridnum=3)
            exp.surrogate.pre_train(exp.surr_pre_it, 0.03, 0.9999, 500, store=True)
        # Load the surrogate
        exp.surrogate.surrogate_load()

        # Define the log density        
        def log_density(x, model, surrogate, transform):

            # Compute transformation log Jacobian
            adjust = transform.compute_log_jacob_func(x)

            # Eval model or surrogate
            if surrogate:
              modelOut = surrogate.forward(x)
            else:
              modelOut = model.solve_t(transform(x))
            stds = model.defOut * model.stdRatio
            Data = torch.tensor(model.data).to(exp.device)

            # Compute LL
            ll1 = -0.5 * np.prod(model.data.shape) * np.log(2.0 * np.pi)  # a number
            ll2 = (-0.5 * model.data.shape[1] * torch.log(torch.prod(stds))).item()  # a number
            ll3 = 0.0
            for i in range(4):
              ll3 += - 0.5 * torch.sum(((modelOut[:, i].unsqueeze(1) - Data[i, :].unsqueeze(0)) / stds[0, i]) ** 2, dim=1)
            negLL = -(ll1 + ll2 + ll3) 
            res = -negLL.reshape(x.size(0), 1) + adjust
            
            # Return LL
            return res
        
        # Assign log-density model
        exp.model_logdensity = lambda x: log_density(x, model, exp.surrogate,trsf)

        # Run VI
        exp.run()


    def rc_example(self):

        if "it" in os.environ:
          max_it  = int(os.environ["it"])
          max_pre = 1000
        else:
          max_it  = 25001
          max_pre = 40000

        print('')
        print('--- TEST 3: RC MODEL - NOFAS')
        print('')

        # Import model
        from linfa.models.circuitModels import rcModel
        
        exp = experiment()
        exp.name = "rc"
        exp.flow_type         = 'maf'         # str: Type of flow (default 'realnvp')
        exp.n_blocks          = 5             # int: Number of layers (default 5)
        exp.hidden_size       = 100           # int: Hidden layer size for MADE in each layer (default 100)
        exp.n_hidden          = 1             # int: Number of hidden layers in each MADE (default 1)
        exp.activation_fn     = 'relu'        # str: Actication function used (default 'relu')
        exp.input_order       = 'sequential'  # str: Input order for create_mask (default 'sequential')
        exp.batch_norm_order  = True          # bool: Order to decide if batch_norm is used (default True)
        exp.save_interval = 5000          # int: How often to sample from normalizing flow

        exp.input_size    = 2       # int: Dimensionality of input (default 2)
        exp.batch_size    = 250     # int: Number of samples generated (default 100)
        exp.true_data_num = 2       # int: number of true model evaluated (default 2)
        # exp.n_iter        = 25001   # int: Number of iterations (default 25001)
        exp.n_iter        = max_it
        exp.lr            = 0.005   # float: Learning rate (default 0.003)
        exp.lr_decay      = 0.9999  # float: Learning rate decay (default 0.9999)
        exp.log_interval  = 10      # int: How often to show loss stat (default 10)

        exp.run_nofas          = True
        exp.surr_pre_it       = max_pre #:int: Number of pre-training iterations for surrogate model
        exp.surr_upd_it       = 6000  #:int: Number of iterations for the surrogate model update

        exp.annealing          = False
        exp.calibrate_interval = 1000  # int: How often to update surrogate model (default 1000)
        exp.budget             = 64    # int: Total number of true model evaluation
        exp.surr_folder        = "./"
        exp.use_new_surr       = True

        exp.output_dir   = './results/' + exp.name
        exp.log_file     = 'log.txt'
        exp.seed         = random.randint(0, 10 ** 9)  # int: Random seed used
        exp.n_sample     = 5000                        # int: Batch size to generate final results/plots
        exp.no_cuda      = True                        # Running on CPU by default but tested on CUDA

        exp.optimizer = 'RMSprop'
        exp.lr_scheduler = 'ExponentialLR'

        exp.device = torch.device('cuda:0' if torch.cuda.is_available() and not exp.no_cuda else 'cpu')

        # Define transformation
        # One list for each variable
        trsf_info = [['tanh',-7.0,7.0,100.0,1500.0],
                     ['exp',-7.0,7.0,1.0e-5,1.0e-2]]
        trsf = Transformation(trsf_info)
        exp.transform = trsf

        # Define model
        cycleTime = 1.07
        totalCycles = 10
        res_path = os.path.abspath(os.path.join(os.path.dirname(linfa.__file__), os.pardir))
        forcing = np.loadtxt(res_path + '/resource/data/inlet.flow')
        model = rcModel(cycleTime, totalCycles, forcing, device=exp.device)  # RCR Model Defined
        exp.model = model

        # Read Data
        model.data = np.loadtxt(res_path + '/resource/data/data_rc.txt')

        # Define surrogate model
        exp.surrogate = Surrogate(exp.name, lambda x: model.solve_t(trsf.forward(x)), exp.input_size, 3,
                                  model_folder=exp.surr_folder, limits=torch.Tensor([[-7, 7], [-7, 7]]), 
                                  memory_len=20, device=exp.device)
        surr_filename = exp.surr_folder + exp.name
        if exp.use_new_surr or (not os.path.isfile(surr_filename + ".sur")) or (not os.path.isfile(surr_filename + ".npz")):
            print("Warning: Surrogate model files: {0}.npz and {0}.npz could not be found. ".format(surr_filename))
            exp.surrogate.gen_grid(gridnum=4)
            exp.surrogate.pre_train(exp.surr_pre_it, 0.03, 0.9999, 500, store=True)
        # Load the surrogate
        exp.surrogate.surrogate_load()

        # Define log density
        def log_density(x, model, surrogate, transform):
            
            # Compute transformation log Jacobian
            adjust = transform.compute_log_jacob_func(x)

            if surrogate:
                modelOut = surrogate.forward(x)
            else:
                modelOut = model.solve_t(transform.forward(x))

            # Get the absolute values of the standard deviations
            stds = model.defOut * model.stdRatio
            Data = torch.tensor(model.data).to(exp.device)
            
            # Eval Gaussian LL
            ll1 = -0.5 * np.prod(model.data.shape) * np.log(2.0 * np.pi)  # a number
            ll2 = (-0.5 * model.data.shape[1] * torch.log(torch.prod(stds))).item()  # a number
            ll3 = 0.0
            for i in range(3):
                ll3 += - 0.5 * torch.sum(((modelOut[:, i].unsqueeze(1) - Data[i, :].unsqueeze(0)) / stds[0, i]) ** 2, dim=1)
            negLL = -(ll1 + ll2 + ll3)
            res = -negLL.reshape(x.size(0), 1) + adjust
            
            # Return LL
            return res

        # Assign log-density
        exp.model_logdensity = lambda x: log_density(x, model, exp.surrogate, trsf)

        # Run VI
        exp.run()

    def rcr_example(self):

        if "it" in os.environ:
          max_it  = int(os.environ["it"])
          max_pre = 1000
        else:
          max_it  = 25001
          max_pre = 50000

        print('')
        print('--- TEST 4: RCR MODEL - NOFAS')
        print('')

        # Import rcr model
        from linfa.models.circuitModels import rcrModel

        exp = experiment()
        exp.name = "rcr"
        exp.flow_type         = 'maf'         # str: Type of flow (default 'realnvp')
        exp.n_blocks          = 15            # int: Number of layers (default 5)
        exp.hidden_size       = 100           # int: Hidden layer size for MADE in each layer (default 100)
        exp.n_hidden          = 1             # int: Number of hidden layers in each MADE (default 1)
        exp.activation_fn     = 'relu'        # str: Actication function used (default 'relu')
        exp.input_order       = 'sequential'  # str: Input order for create_mask (default 'sequential')
        exp.batch_norm_order  = True          # bool: Order to decide if batch_norm is used (default True)
        exp.save_interval = 5000          # int: How often to sample from normalizing flow

        exp.input_size    = 3       # int: Dimensionality of input (default 2)
        exp.batch_size    = 500     # int: Number of samples generated (default 100)
        exp.true_data_num = 2       # double: number of true model evaluated (default 2)
        # exp.n_iter        = 25001   # int: Number of iterations (default 25001)
        exp.n_iter        = max_it
        exp.lr            = 0.003   # float: Learning rate (default 0.003)
        exp.lr_decay      = 0.9999  # float: Learning rate decay (default 0.9999)
        exp.log_interval  = 10      # int: How often to show loss stat (default 10)

        exp.run_nofas          = True
        exp.surr_pre_it       = max_pre #:int: Number of pre-training iterations for surrogate model
        exp.surr_upd_it       = 6000  #:int: Number of iterations for the surrogate model update

        exp.annealing          = False
        exp.calibrate_interval = 300  # int: How often to update surrogate model (default 1000)
        exp.budget             = 216  # int: Total number of true model evaluation
        exp.surr_folder        = "./"
        exp.use_new_surr       = True

        exp.output_dir = './results/' + exp.name
        exp.log_file = 'log.txt'
        exp.seed = random.randint(0, 10 ** 9)  # int: Random seed used
        exp.n_sample = 5000                    # int: Batch size to generate final results/plots
        exp.no_cuda = True                     # Running on CPU by default but tested on CUDA

        exp.optimizer = 'RMSprop'
        exp.lr_scheduler = 'ExponentialLR'

        exp.device = torch.device('cuda:0' if torch.cuda.is_available() and not exp.no_cuda else 'cpu')

        # Define transformation
        # One list for each variable
        trsf_info = [['tanh',-7.0,7.0,100.0,1500.0],
                     ['tanh',-7.0,7.0,100.0,1500.0],
                     ['exp',-7.0,7.0,1.0e-5,1.0e-2]]
        trsf = Transformation(trsf_info)
        exp.transform = trsf

        # Define model
        cycleTime = 1.07
        totalCycles = 10
        res_path = os.path.abspath(os.path.join(os.path.dirname(linfa.__file__), os.pardir))
        forcing = np.loadtxt(res_path + '/resource/data/inlet.flow')
        model = rcrModel(cycleTime, totalCycles, forcing, device=exp.device)  # RCR Model Defined
        exp.model = model

        # Read data
        model.data = np.loadtxt(res_path + '/resource/data/data_rcr.txt')

        # Define surrogate
        exp.surrogate = Surrogate(exp.name, lambda x: model.solve_t(trsf.forward(x)), exp.input_size, 3,
                                  model_folder=exp.surr_folder, limits=torch.Tensor([[-7, 7], [-7, 7], [-7, 7]]), 
                                  memory_len=20, device=exp.device)
        surr_filename = exp.surr_folder + exp.name
        if exp.use_new_surr or (not os.path.isfile(surr_filename + ".sur")) or (not os.path.isfile(surr_filename + ".npz")):
            print("Warning: Surrogate model files: {0}.npz and {0}.npz could not be found. ".format(surr_filename))
            exp.surrogate.gen_grid(gridnum=4)
            exp.surrogate.pre_train(exp.surr_pre_it, 0.03, 0.9999, 500, store=True)
        # Load the surrogate
        exp.surrogate.surrogate_load()

        # Define log density
        def log_density(x, model, surrogate, transform):

            # Compute transformation log Jacobian
            adjust = transform.compute_log_jacob_func(x)

            if surrogate:
                modelOut = surrogate.forward(x)
            else:
                modelOut = model.solve_t(transform.forward(x))

            # Get the absolute values of the standard deviations
            stds = model.defOut * model.stdRatio
            Data = torch.tensor(model.data).to(exp.device)
            
            # Eval LL
            ll1 = -0.5 * np.prod(model.data.shape) * np.log(2.0 * np.pi)  # a number
            ll2 = (-0.5 * model.data.shape[1] * torch.log(torch.prod(stds))).item()  # a number
            ll3 = 0.0
            for i in range(3):
                ll3 += - 0.5 * torch.sum(((modelOut[:, i].unsqueeze(1) - Data[i, :].unsqueeze(0)) / stds[0, i]) ** 2, dim=1)
            negLL = -(ll1 + ll2 + ll3)
            res = -negLL.reshape(x.size(0), 1) + adjust

            # Return LL
            return res

        # Assign logdensity model
        exp.model_logdensity = lambda x: log_density(x, model, exp.surrogate, trsf)

        # Run VI
        exp.run()


    def adaann_example(self):

        # Quick run when repository is pushed
        if "it" in os.environ:
          run_T_1  = int(os.environ["it"])
          run_T    = 1
          run_T_0  = 1
          run_t0   = 0.99          
        else:
          run_T_1  = 10000
          run_T    = 3
          run_T_0  = 500
          run_t0   = 0.001          

        print('')
        print('--- TEST 5: FRIEDMAN 1 MODEL - ADAANN')
        print('')

        from linfa.run_experiment import experiment

        # Experiment Setting
        exp = experiment()
        exp.name              = "adaann"
        exp.flow_type         = 'realnvp'     # str: Type of flow (default 'realnvp')
        exp.n_blocks          = 10            # int: Number of layers (default 5)
        exp.hidden_size       = 20            # int: Hidden layer size for MADE in each layer (default 100)
        exp.n_hidden          = 1             # int: Number of hidden layers in each MADE (default 1)
        exp.activation_fn     = 'relu'        # str: Actication function used (default 'relu')
        exp.input_order       = 'sequential'  # str: Input order for create_mask (default 'sequential')
        exp.batch_norm_order  = True          # bool: Order to decide if batch_norm is used (default True)
        exp.save_interval     = 1000          # int: How often to sample from normalizing flow

        exp.input_size   = 10     # int: Dimensionality of input (default 2)
        exp.batch_size   = 100    # int: Number of samples generated (default 100)
        exp.lr           = 0.005  # float: Learning rate (default 0.003)
        # exp.lr_decay   = 0.75
        exp.lr_decay     = 0.9
        exp.log_interval = 10     # int: How often to show loss stat (default 10)
        exp.run_nofas    = False
        exp.annealing    = True

        exp.optimizer    = 'Adam'    # str: type of optimizer used
        exp.lr_scheduler = 'StepLR'  # str: type of lr scheduler used
        exp.lr_step      = 500       # int: number of steps for lr step scheduler
        exp.tol          = 0.4       # float: tolerance for AdaAnn scheduler
        # exp.t0           = 0.001     # float: initial inverse temperature value
        exp.t0           = run_t0
        exp.N            = 100       # int: number of sample points during annealing
        exp.N_1          = 400       # int: number of sample points at t=1
        # exp.T_0          = 500       # int: number of parameter updates at initial t0
        # exp.T            = 3         # int: number of parameter updates during annealing
        # exp.T_1          = 10000     # int: number of parameter updates at t=1
        exp.T_0          = run_T_0
        exp.T            = run_T
        exp.T_1          = run_T_1
        exp.M            = 1000      # int: number of sample points used to update temperature
        exp.scheduler    = 'AdaAnn'  # str: type of annealing scheduler used

        exp.output_dir   = './results/' + exp.name
        exp.log_file     = 'log.txt'
        # exp.seed = 35435  # int: Random seed used
        exp.seed         = random.randint(0, 10 ** 9)  # int: Random seed used
        exp.n_sample     = 5000                        # int: Batch size to generate final results/plots
        exp.no_cuda      = True                        # Running on CPU by default but tested on CUDA

        exp.device = torch.device('cuda:0' if torch.cuda.is_available() and not exp.no_cuda else 'cpu')

        # Model Setting
        res_path = os.path.abspath(os.path.join(os.path.dirname(linfa.__file__), os.pardir))
        data_set = np.loadtxt(res_path + '/resource/data/D1000.csv',delimiter=',',skiprows=1)
        data = torch.tensor(data_set).to(exp.device)

        def log_density(params, d):

            # Compute Model
            # Modified bimodal version of the Friedman model
            def targetPosterior(b, x):
                return b[0] * torch.sin(np.pi * x[:, 0] * x[:, 1]) + \
                       (b[1] ** 2) * (x[:, 2] - b[2]) ** 2 + \
                       x[:, 3] * b[3] + \
                       x[:, 4] * b[4] + \
                       x[:, 5] * b[5] + \
                       x[:, 6] * b[6] + \
                       x[:, 7] * b[7] + \
                       x[:, 8] * b[8] + \
                       x[:, 9] * b[9]

            f = torch.zeros(len(params)).to(exp.device)

            for i in range(len(params)):
                y_out = targetPosterior(params[i], d)
                val = torch.linalg.norm(y_out - d[:, 10])
                f[i] = -val ** 2 / 2

            return f

        exp.model_logdensity = lambda x: log_density(x, data)
        exp.run()
    
    def rcr_nofas_adaann_example(run_nofas=True, run_adaann=False):
        
        # Quick run when repository is pushed
        if "it" in os.environ:
          max_it  = int(os.environ["it"])
          max_pre = 1000
        else:
          max_it  = 25001
          max_pre = 50000

        print('')
        print('--- TEST 6: RCR MODEL - NOFAS and ADAANN')
        print('')

        # Import rcr model
        from linfa.models.circuitModels import rcrModel

        exp = experiment()
        exp.name = "rcr_nofas_adaann"
        exp.flow_type         = 'maf'         # str: Type of flow (default 'realnvp')
        exp.n_blocks          = 15            # int: Number of layers (default 5)
        exp.hidden_size       = 100           # int: Hidden layer size for MADE in each layer (default 100)
        exp.n_hidden          = 1             # int: Number of hidden layers in each MADE (default 1)
        exp.activation_fn     = 'relu'        # str: Actication function used (default 'relu')
        exp.input_order       = 'sequential'  # str: Input order for create_mask (default 'sequential')
        exp.batch_norm_order  = True          # bool: Order to decide if batch_norm is used (default True)
        exp.save_interval     = 200           # int: How often to sample from normalizing flow

        exp.input_size    = 3       # int: Dimensionality of input (default 2)
        exp.batch_size    = 500     # int: Number of samples generated (default 100)
        exp.true_data_num = 2       # double: number of true model evaluated (default 2)
        # exp.n_iter        = 25001   # int: Number of iterations (default 25001)
        exp.n_iter        = max_it
        exp.lr            = 0.003   # float: Learning rate (default 0.003)
        exp.lr_decay      = 0.9999  # float: Learning rate decay (default 0.9999)
        exp.log_interval  = 10      # int: How often to show loss stat (default 10)

        exp.run_nofas          = True
        exp.calibrate_interval = 300  # int: How often to update surrogate model (default 1000)
        exp.budget             = 216  # int: Total number of true model evaluation
        exp.surr_pre_it       = max_pre #:int: Number of pre-training iterations for surrogate model
        exp.surr_upd_it       = 6000  #:int: Number of iterations for the surrogate model update
        exp.surr_folder        = "./"
        exp.use_new_surr       = True

        exp.annealing = True
        exp.tol       = 0.01     # float: tolerance for AdaAnn scheduler
        exp.t0        = 0.05     # float: initial inverse temperature value
        exp.N         = 100      # int: number of sample points during annealing
        exp.N_1       = 100      # int: number of sample points at t=1
        exp.T_0       = 500      # int: number of parameter updates at initial t0
        exp.T         = 5        # int: number of parameter updates during annealing
        exp.T_1       = 5000     # int: number of parameter updates at t=1
        exp.M         = 1000     # int: number of sample points used to update temperature
        exp.scheduler = 'AdaAnn' # str: type of annealing scheduler used
    
        exp.output_dir = './results/' + exp.name
        exp.log_file = 'log.txt'
        exp.seed = random.randint(0, 10 ** 9)  # int: Random seed used
        exp.n_sample = 5000                    # int: Batch size to generate final results/plots
        exp.no_cuda = False                     # Running on CPU by default but tested on CUDA

        exp.optimizer = 'RMSprop'
        exp.lr_scheduler = 'ExponentialLR'

        exp.device = torch.device('cuda:0' if torch.cuda.is_available() and not exp.no_cuda else 'cpu')

        # Define transformation
        # One list for each variable
        trsf_info = [['tanh',-7.0,7.0,100.0,1500.0],
                     ['tanh',-7.0,7.0,100.0,1500.0],
                     ['exp',-7.0,7.0,1.0e-5,1.0e-2]]
        trsf = Transformation(trsf_info)
        exp.transform = trsf

        # Define model
        cycleTime = 1.07
        totalCycles = 10
        res_path = os.path.abspath(os.path.join(os.path.dirname(linfa.__file__), os.pardir))
        forcing = np.loadtxt(res_path + '/resource/data/inlet.flow')
        model = rcrModel(cycleTime, totalCycles, forcing, device=exp.device)  # RCR Model Defined
        exp.model = model

        # Read data
        model.data = np.loadtxt(res_path + '/resource/data/data_rcr.txt')

        # Define surrogate
        exp.surrogate = Surrogate(exp.name, lambda x: model.solve_t(trsf.forward(x)), exp.input_size, 3,
                                  model_folder=exp.surr_folder, limits=torch.Tensor([[-7, 7], [-7, 7], [-7, 7]]), 
                                  memory_len=20, device=exp.device)
        surr_filename = exp.surr_folder + exp.name
        if exp.use_new_surr or (not os.path.isfile(surr_filename + ".sur")) or (not os.path.isfile(surr_filename + ".npz")):
            print("Warning: Surrogate model files: {0}.npz and {0}.npz could not be found. ".format(surr_filename))
            exp.surrogate.gen_grid(gridnum=4)
            exp.surrogate.pre_train(exp.surr_pre_it, 0.03, 0.9999, 500, store=True)
        # Load the surrogate
        exp.surrogate.surrogate_load()

        # Define log density
        def log_density(x, model, surrogate, transform):

            # Compute transformation log Jacobian
            adjust = transform.compute_log_jacob_func(x)

            if surrogate:
                modelOut = surrogate.forward(x)
            else:
                modelOut = model.solve_t(transform.forward(x))

            # Get the absolute values of the standard deviations
            stds = model.defOut * model.stdRatio
            Data = torch.tensor(model.data).to(exp.device)
            
            # Eval LL
            ll1 = -0.5 * np.prod(model.data.shape) * np.log(2.0 * np.pi)  # a number
            ll2 = (-0.5 * model.data.shape[1] * torch.log(torch.prod(stds))).item()  # a number
            ll3 = 0.0
            for i in range(3):
                ll3 += - 0.5 * torch.sum(((modelOut[:, i].unsqueeze(1) - Data[i, :].unsqueeze(0)) / stds[0, i]) ** 2, dim=1)
            negLL = -(ll1 + ll2 + ll3)
            res = -negLL.reshape(x.size(0), 1) + adjust

            # Return LL
            return res

        # Assign logdensity model
        exp.model_logdensity = lambda x: log_density(x, model, exp.surrogate, trsf)

        # Run VI
        exp.run()


if __name__ == '__main__':
    
    # Execute tests
    unittest.main()



import os
import torch
import numpy as np
from linfa.maf import MAF, RealNVP

torch.set_default_tensor_type(torch.DoubleTensor)

class experiment:
    """Defines an instance of variational inference

    This class is the core class of the LINFA library
    and defines all the default hyperparameter values and 
    and functions used for inference. 
    """
    def __init__(self):

        # NF ARCHITECTURE parameters
        self.name              = "Experiment"
        self.input_size        = 3             #:int:  Number of input parameters
        self.flow_type         = 'maf'         #:str:  Type of flow ('maf' or 'realnvp')
        self.n_blocks          = 15            #:int:  Number of layers
        self.hidden_size       = 100           #:int:  Hidden layer size for MADE in each layer
        self.n_hidden          = 1             #:int:  Number of hidden layers in each MADE
        self.activation_fn     = 'relu'        #:str:  Actication function used (either 'relu','tanh' or 'sigmoid')
        self.input_order       = 'sequential'  #:str:  Input order for create_mask (either 'sequential' or 'random')
        self.batch_norm_order  = True          #:bool: Uses decide if batch_norm is used

        # NOFAS parameters
        self.run_nofas          = True  #:bool:   Activate NoFAS and the use of a surrogate model
        self.log_interval       = 10    #:int:    How often the loss statistics are printed
        self.calibrate_interval = 300   #:int:    How often the surrogate model is updated
        self.true_data_num      = 2     #:double: Number of true model evaluated at each surrogate update
        self.budget             = 216   #:int:    Maximum number of allowed evaluations of the true model
        self.surr_pre_it        = 40000 #:int:    Number of pre-training iterations for surrogate model
        self.surr_upd_it        = 6000  #:int:    Number of iterations for the surrogate model update
        self.surr_folder        = "./"  #:str:    Folder where the surrogate model is stored
        self.use_new_surr       = True #:bool:   Start by pre-training a new surrogate and ignore existing surrogates

        # OPTIMIZER parameters
        self.optimizer    = 'Adam'   #:str:    Type of optimizer used (either 'Adam' or 'RMSprop')
        self.lr           = 0.003    #:double: Learning rate
        self.lr_decay     = 0.9999   #:double: Learning rate decay
        self.lr_scheduler = 'StepLR' #:str:    type of lr scheduler used (either 'StepLR' or 'ExponentialLR')
        self.lr_step      = 1000     #:int:    Number of steps for StepLR learning rate scheduler 
        self.batch_size   = 500      #:int:    Number of batch samples generated at every iteration from the base distribution        
        self.n_iter       = 25001    #:int:    Total number of iterations

        # ANNEALING parameters
        self.annealing    = True     #:bool:   Flag to activate an annealing scheduler
        self.scheduler    = 'AdaAnn' #:str:    Type of annealing scheduler (either 'AdaAnn' or 'Linear')
        # AdaAnn
        self.tol          = 0.001    #:double: KL tolerance for AdaAnn scheduler
        self.t0           = 0.01     #:double: Initial value for the inverse temperature
        self.N            = 100      #:int:    Number of batch samples generated for $t<1$ at each iteration
        self.N_1          = 1000     #:int:    number of batch samples generated for $t=1$ at each iteration
        self.T_0          = 500      #:int:    Number of parameter updates at the initial inverse temperature $t_0$
        self.T            = 5        #:int:    Number of parameter updates for each temperature for $t<1$
        self.T_1          = 5001     #:int:    Number of parameter updates at $t=1$
        self.M            = 1000     #:int:    Number of Monte Carlo  samples use to compute the denominator of the AdaAnn formula        
        # Linear scheduler
        self.linear_step  = 0.0001   #:double: Fixed step size for the Linear annealing scheduler

        # OUTPUT parameters
        self.output_dir          = './results/' + self.name #:str: Name of the output folder
        self.log_file            = 'log.txt'                #:str: File name where the log profile stats are written
        self.seed                = 35435                    #:int: Random seed
        self.n_sample            = 5000                     #:int: Number of batch samples used to print results at save_interval
        self.save_interval       = 200                      #:int: Save interval for all results
        self.store_nf_interval   = 1000                     #:int: Save interval for normalizing flow parameters
        self.store_surr_interval = None                     #:int: Save interval for surrogate model (None for no save)

        # DEVICE parameters
        self.no_cuda = True #:bool: Flag to use CPU

        # Set device
        self.device = torch.device('cuda:0' if torch.cuda.is_available() and not self.no_cuda else 'cpu')

        # Local pointer to the main components for inference
        self.transform        = None
        self.model            = None
        self.model_logdensity = None
        self.surrogate        = None

    def run(self):
        """Runs instance of inference inference problem        

        """

        # Check is surrogate exists
        if self.run_nofas:
            if not os.path.exists(self.name + ".sur") or not os.path.exists(self.name + ".npz"):
                print("Abort: NoFAS enabled, without surrogate files. \nPlease include the following surrogate files in root directory.\n{}.sur and {}.npz".format(self.name, self.name))
                exit(0)
        # setup file ops
        if not os.path.isdir(self.output_dir):
            os.makedirs(self.output_dir)

        # Save a copy of the data in the result folder so it is handy
        if hasattr(self.model,'data'):
          np.savetxt(self.output_dir + '/' + self.name + '_data', self.model.data, newline="\n")

        # setup device
        torch.manual_seed(self.seed)
        if self.device.type == 'cuda': torch.cuda.manual_seed(self.seed)

        print('')
        print('--- Running on device: '+ str(self.device))
        print('')

        # model
        if self.flow_type == 'maf':
            nf = MAF(self.n_blocks, self.input_size, self.hidden_size, self.n_hidden, None,
                     self.activation_fn, self.input_order, batch_norm=self.batch_norm_order)
        elif self.flow_type == 'realnvp':  # Under construction
            nf = RealNVP(self.n_blocks, self.input_size, self.hidden_size, self.n_hidden, None,
                         batch_norm=self.batch_norm_order)
        else:
            raise ValueError('Unrecognized model.')

        nf = nf.to(self.device)
        if self.optimizer == 'Adam':
            optimizer = torch.optim.Adam(nf.parameters(), lr=self.lr)
        elif self.optimizer == 'RMSprop':
            optimizer = torch.optim.RMSprop(nf.parameters(), lr=self.lr)
        else:
            raise ValueError('Unrecognized optimizer.')

        if self.lr_scheduler == 'StepLR':
            scheduler = torch.optim.lr_scheduler.StepLR(optimizer, self.lr_step, self.lr_decay)
        elif self.lr_scheduler == 'ExponentialLR':
            scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, self.lr_decay)
        else:
            raise ValueError('Unrecognized learning rate scheduler.')

        if self.annealing:
            i = 1
            prev_i = 1
            tvals = []
            t = self.t0
            dt = 0.0
            loglist = []
            while t < 1:
                t = min(1, t + dt)
                tvals = np.concatenate([tvals, np.array([t])])
                self.n_iter = self.T
                self.batch_size = self.N
                
                if t == self.t0:
                    self.n_iter = self.T_0
                if t == 1:
                    self.batch_size = self.N_1
                    self.n_iter = self.T_1

                while i < prev_i + self.n_iter:
                    self.train(nf, optimizer, i, loglist, sampling=True, t=t)
                    if t == 1:
                        scheduler.step()
                    i += 1
                prev_i = i

                if self.scheduler == 'AdaAnn':
                    z0 = nf.base_dist.sample([self.M])
                    zk, _ = nf(z0)
                    log_qk = self.model_logdensity(zk)
                    dt = self.tol / torch.sqrt(log_qk.var())
                    dt = dt.detach()# .numpy()

                if self.scheduler == 'Linear':
                    dt = self.linear_step
        else:
            loglist = []
            for i in range(1, self.n_iter+1):                
                self.train(nf, optimizer, i, loglist, sampling=True)
                scheduler.step()

        print('')
        print('--- Simulation completed!')        

    def train(self, nf, optimizer, iteration, log, sampling=True, t=1):
        """Parameter update for normalizing flow and surrogate model

        This is the function where the ELBO loss function is evaluated, 
        the results are saved and the surrogate model is updated.
        
        Args:
            nf (instance of normalizing flow): the normalizing flow architecture used for variational inference
            optimizer (instance of PyTorch optimizer): the selected PyTorch optimizer
            iteration (int): current iteration number
            log (list of lists): stores a log of [iteration, annealing temperature, loss value]
            sampling (bool): Flag indicating the sampling stage
            t (double): current inverse temperature for annealing scheduler

        Returns:
            None
        """

        # Set the normalizing flow in training mode
        nf.train()

        # Evaluate the Jacobian terms in  loss function
        x0 = nf.base_dist.sample([self.batch_size])
        xk, sum_log_abs_det_jacobians = nf(x0)

        # generate and save samples evaluation
        if sampling and iteration % self.save_interval == 0:
            print('--- Saving results at iteration '+str(iteration))
            x00 = nf.base_dist.sample([self.n_sample])
            xkk, _ = nf(x00)
            # Save surrogate grid
            if not(self.surrogate is None):
              np.savetxt(self.output_dir + '/' + self.name + '_grid_' + str(iteration), self.surrogate.grid_record.clone().cpu().numpy(), newline="\n")
            # Save log profile
            np.savetxt(self.output_dir + '/' + self.log_file, np.array(log), newline="\n")
            # Save transformed samples          
            np.savetxt(self.output_dir + '/' + self.name + '_samples_' + str(iteration), xkk.data.clone().cpu().numpy(), newline="\n")
            # Save samples in the original space
            if not(self.transform is None):
              xkk_samples = self.transform.forward(xkk).data.cpu().numpy()
              np.savetxt(self.output_dir + '/' + self.name + '_params_' + str(iteration), xkk_samples, newline="\n")
            else:
              xkk_samples = xkk.data.cpu().numpy()
              np.savetxt(self.output_dir + '/' + self.name + '_params_' + str(iteration), xkk_samples, newline="\n")
            # Save marginal statistics
            np.savetxt(self.output_dir + '/' + self.name + '_marginal_stats_' + str(iteration), np.concatenate((xkk_samples.mean(axis=0).reshape(-1,1),xkk_samples.std(axis=0).reshape(-1,1)),axis=1), newline="\n")
            # Save log density at the same samples
            np.savetxt(self.output_dir + '/' + self.name + '_logdensity_' + str(iteration), self.model_logdensity(xkk).data.cpu().numpy(), newline="\n")
            # Save model outputs at the samples - If a model is defined
            if not(self.transform is None):
              stds = torch.abs(self.model.solve_t(self.model.defParam)).to(self.device) * self.model.stdRatio
              o00 = torch.randn(x00.size(0), self.model.data.shape[0]).to(self.device)
              noise = o00*stds.repeat(o00.size(0),1)
              if self.surrogate:
                np.savetxt(self.output_dir + '/' + self.name + '_outputs_' + str(iteration), (self.surrogate.forward(xkk) + noise).data.cpu().numpy(), newline="\n")
              else:
                np.savetxt(self.output_dir + '/' + self.name + '_outputs_' + str(iteration), (self.model.solve_t(self.transform.forward(xkk)) + noise).data.cpu().numpy(), newline="\n")

        if torch.any(torch.isnan(xk)):
            print("Error: samples xk are nan at iteration " + str(iteration))
            print(xk)
            np.savetxt(self.output_dir + '/' + self.log_file, np.array(log), newline="\n")
            exit(-1)

        # updating surrogate model
        if self.run_nofas and iteration % self.calibrate_interval == 0 and self.surrogate.grid_record.size(0) < self.budget:
            xk0 = xk[:self.true_data_num, :].data.clone()            
            # print("\n")
            # print(list(self.surrogate.grid_record.size())[0])
            # print(xk0)
            self.surrogate.update(xk0, max_iters=self.surr_upd_it)

        # Free energy bound
        loss = (- torch.sum(sum_log_abs_det_jacobians, 1) - t * self.model_logdensity(xk)).mean()
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        if iteration % self.log_interval == 0:
            print('VI NF (t=%5.3f): it: %7d | loss: %8.3e' % (t,iteration, loss.item()))
            log.append([t, iteration, loss.item()])
        
        # Save state of normalizing flow layers
        if self.store_nf_interval > 0 and iteration % self.store_nf_interval == 0:
            torch.save(nf.state_dict(), self.output_dir + '/' + self.name + "_" + str(iteration) + ".nf")

        if not(self.store_surr_interval is None) and self.store_surr_interval > 0 and iteration % self.store_surr_interval == 0:
            self.surrogate.surrogate_save() # Save surrogate model